import uuid

import pytest
from unittest.mock import AsyncMock, MagicMock
from dawnet_client import WebSocketClient, DAWNetFilePath
from dawnet_client import ui_param


# Example method to register
async def example_method_one(a: int, b: float, c: str, d: DAWNetFilePath):
    pass

@pytest.mark.asyncio
async def test_register_method_one():
    # SETUP
    client = WebSocketClient('127.0.0.1', '1234')
    client.connect = AsyncMock()
    client.set_token(str(uuid.uuid4()))

    # EXECUTE
    await client.register_method(example_method_one)

    # ASSERTS
    client.connect.assert_called_once()
    assert 'example_method_one' in client.method_details
    assert client.method_details['example_method_one']['params'] == [
        {'name': 'a', 'type': 'int', 'default_value': None, 'ui_component': None},
        {'name': 'b', 'type': 'float', 'default_value': None, 'ui_component': None},
        {'name': 'c', 'type': 'str', 'default_value': None, 'ui_component': None},
        {'name': 'd', 'type': 'DAWNetFilePath', 'default_value': None, 'ui_component': None}]

async def example_method_one_defaults(a: int=5, b: float=2.2, c: str='hello', d: DAWNetFilePath=None):
    pass

@pytest.mark.asyncio
async def test_register_method_one_defaults():
    # SETUP
    client = WebSocketClient('127.0.0.1', '1234')
    client.connect = AsyncMock()
    client.set_token(str(uuid.uuid4()))

    # EXECUTE
    await client.register_method(example_method_one_defaults)

    # ASSERTS
    client.connect.assert_called_once()
    assert 'example_method_one_defaults' in client.method_details
    assert client.method_details['example_method_one_defaults']['params'] == [
        {'name': 'a', 'type': 'int', 'default_value': 5, 'ui_component': None},
        {'name': 'b', 'type': 'float', 'default_value': 2.2, 'ui_component': None},
        {'name': 'c', 'type': 'str', 'default_value': 'hello', 'ui_component': None},
        {'name': 'd', 'type': 'DAWNetFilePath', 'default_value': None, 'ui_component': None}
    ]

async def example_method_one_partial_defaults(a: int, b: float, c: str='hello', d: DAWNetFilePath=None):
    pass

@pytest.mark.asyncio
async def test_method_one_partial_defaults():
    # SETUP
    client = WebSocketClient('127.0.0.1', '1234')
    client.connect = AsyncMock()
    client.set_token(str(uuid.uuid4()))

    # EXECUTE
    await client.register_method(example_method_one_partial_defaults)

    # ASSERTS
    client.connect.assert_called_once()
    assert 'example_method_one_partial_defaults' in client.method_details
    assert client.method_details['example_method_one_partial_defaults']['params'] == [
        {'name': 'a', 'type': 'int', 'default_value': None, 'ui_component': None},
        {'name': 'b', 'type': 'float', 'default_value': None, 'ui_component': None},
        {'name': 'c', 'type': 'str', 'default_value': 'hello', 'ui_component': None},
        {'name': 'd', 'type': 'DAWNetFilePath', 'default_value': None, 'ui_component': None}
    ]

@ui_param('a', 'DAWNetNumberSlider', min=0, max=10, step=1, default=5)
async def example_method_one_with_decorators(a: int, b: float, c: str='hello', d: DAWNetFilePath=None):
    pass

@pytest.mark.asyncio
async def test_register_method_one_with_decorators():
    # SETUP
    client = WebSocketClient('127.0.0.1', '1234')
    client.connect = AsyncMock()
    client.set_token(str(uuid.uuid4()))

    # EXECUTE
    await client.register_method(example_method_one_with_decorators)

    # ASSERTS
    client.connect.assert_called_once()
    assert 'example_method_one_with_decorators' in client.method_details
    assert client.method_details['example_method_one_with_decorators']['params'] == [
        {'name': 'a', 'type': 'int', 'default_value': 5, 'min': 0, 'max': 10, 'step': 1, 'ui_component': 'DAWNetNumberSlider'},
        {'name': 'b', 'type': 'float', 'default_value': None, 'ui_component': None},
        {'name': 'c', 'type': 'str', 'default_value': 'hello', 'ui_component': None},
        {'name': 'd', 'type': 'DAWNetFilePath', 'default_value': None, 'ui_component': None}
    ]

@ui_param('a', 'DAWNetNumberSlider', min=0, max=10, step=1, default=5)
@ui_param('b', 'DAWNetNumberSlider', min=0.0, max=10.0, step=0.5, default=7.7)
async def example_method_one_with_multiple_decorators(a: int, b: float=7.7, c: str='hello', d: DAWNetFilePath=None):
    pass

@pytest.mark.asyncio
async def test_register_method_one_with_multiple_decorators():
    # SETUP
    client = WebSocketClient('127.0.0.1', '1234')
    client.connect = AsyncMock()
    client.set_token(str(uuid.uuid4()))

    # EXECUTE
    await client.register_method(example_method_one_with_multiple_decorators)

    # ASSERTS
    client.connect.assert_called_once()
    assert 'example_method_one_with_multiple_decorators' in client.method_details
    assert client.method_details['example_method_one_with_multiple_decorators']['params'] == [
        {'name': 'a', 'type': 'int', 'default_value': 5, 'min': 0, 'max': 10, 'step': 1, 'ui_component': 'DAWNetNumberSlider'},
        {'name': 'b', 'type': 'float', 'default_value': 7.7, 'min': 0.0, 'max': 10.0, 'step': 0.5, 'ui_component': 'DAWNetNumberSlider'},
        {'name': 'c', 'type': 'str', 'default_value': 'hello', 'ui_component': None},
        {'name': 'd', 'type': 'DAWNetFilePath', 'default_value': None, 'ui_component': None}
    ]

@ui_param('a', 'unsupported_ui_component', min=0, max=10, step=1, default=5)
async def example_method_one_with_unsupported_decorators(a: int, b: float, c: str='hello', d: DAWNetFilePath=None):
    pass

@pytest.mark.asyncio
async def test_register_method_one_with_unsupported_decorators():
    # SETUP
    client = WebSocketClient('127.0.0.1', '1234')
    client.connect = AsyncMock()
    client.set_token(str(uuid.uuid4()))

    # EXECUTE and ASSERT
    with pytest.raises(ValueError) as exc_info:
        await client.register_method(example_method_one_with_unsupported_decorators)

    assert "Unsupported UI component" in str(exc_info.value)


@ui_param('a', 'DAWNetNumberSlider', min=0, max=10, fakeparam='fake', step=1, default=5)
async def example_method_one_with_unsupported_param(a: int, b: float, c: str='hello', d: DAWNetFilePath=None):
    pass

@pytest.mark.asyncio
async def test_register_method_one_with_unsupported_param():
    # SETUP
    client = WebSocketClient('127.0.0.1', '1234')
    client.connect = AsyncMock()
    client.set_token(str(uuid.uuid4()))

    # EXECUTE and ASSERT
    with pytest.raises(ValueError) as exc_info:
        await client.register_method(example_method_one_with_unsupported_param)

    assert "Unsupported UI param" in str(exc_info.value)


@ui_param('a', 'DAWNetNumberSlider', min=0, max=10, default=5)
async def example_method_one_missing_required_param(a: int, b: float, c: str='hello', d: DAWNetFilePath=None):
    pass

@pytest.mark.asyncio
async def test_register_method_one_missing_required_param():
    # SETUP
    client = WebSocketClient('127.0.0.1', '1234')
    client.connect = AsyncMock()
    client.set_token(str(uuid.uuid4()))

    # EXECUTE and ASSERT
    with pytest.raises(ValueError) as exc_info:
        await client.register_method(example_method_one_missing_required_param)

    assert  "Missing required param(s)" in str(exc_info.value)


@ui_param('c', 'DAWNetMultiChoice', options=['one', 'two', 'three'], default='two')
async def example_method_multichoice(a: int, b: float, c: str, d: DAWNetFilePath):
    pass

@pytest.mark.asyncio
async def test_register_method_example_method_multichoice():
    # SETUP
    client = WebSocketClient('127.0.0.1', '1234')
    client.connect = AsyncMock()
    client.set_token(str(uuid.uuid4()))

    # EXECUTE
    await client.register_method(example_method_multichoice)

    # ASSERTS
    client.connect.assert_called_once()
    assert 'example_method_multichoice' in client.method_details
    assert client.method_details['example_method_multichoice']['params'] == [
        {'name': 'a', 'type': 'int', 'ui_component': None, 'default_value': None},
        {'name': 'b', 'type': 'float', 'ui_component': None, 'default_value': None},
        {'name': 'c', 'type': 'str', 'options': ['one','two','three'], 'default_value': 'two', 'ui_component': 'DAWNetMultiChoice'},
        {'name': 'd', 'type': 'DAWNetFilePath', 'default_value': None, 'ui_component': None}
    ]


# decorators.py
def ui_param(name, ui_component, **kwargs):
    def decorator(func):
        if not hasattr(func, '_ui_params'):
            func._ui_params = {}
        func._ui_params[name] = {'ui_component': ui_component, **kwargs}
        return func
    return decorator



# Existing imports...
import json
import os
import subprocess
from ..dn_tracer import SentryEventLogger, DNSystemType, DNMsgStage, DNTag
from ..file_uploader import FileUploader
from ..utils.audio_utils import process_audio_file


# ResultsHandler class to handle the results
class ResultsHandler:
    def __init__(self, websocket, token, target_sample_rate=41000, target_bit_depth=16, target_channels=2,
                 target_format="wav"):
        self.websocket = websocket
        self.token = token
        self.message_id = None
        self.errors = []
        self.files = []
        self.logs = ""
        self.messages = []
        self.file_uploader = FileUploader()
        self.dn_tracer = SentryEventLogger(DNSystemType.DN_CLIENT.value)

        # Check if ffmpeg is installed
        self.ffmpeg_installed = self.check_ffmpeg()

        # Default target audio settings
        self.target_sample_rate = target_sample_rate
        self.target_bit_depth = target_bit_depth
        self.target_channels = target_channels
        self.target_format = target_format

    def check_ffmpeg(self):
        try:
            subprocess.run(["ffmpeg", "-version"], stdout=subprocess.DEVNULL, stderr=subprocess.STDOUT)
            return True
        except FileNotFoundError:
            print("ffmpeg is not installed. Please install ffmpeg for audio processing.")
            return False

    def update_token(self, token):
        self.token = token

    async def add_error(self, error):
        self.errors.append(error)
        return True

    def set_message_id(self, message_id):
        self.message_id = message_id

    async def add_file(self, file_path):

        if not self.ffmpeg_installed:
            error_message = ("FFmpeg is not installed, which is required for processing audio files.\n"
                             "To install FFmpeg, follow these instructions:\n"
                             "- On macOS: Use Homebrew by running 'brew install ffmpeg' in the terminal.\n"
                             "- On Linux (Debian/Ubuntu): Run 'sudo apt-get install ffmpeg' in the terminal.\n"
                             "- On Linux (Fedora): Run 'sudo dnf install ffmpeg' in the terminal.\n"
                             "- On Linux (Arch Linux): Run 'sudo pacman -S ffmpeg' in the terminal.\n"
                             "For other operating systems or more detailed instructions, visit the FFmpeg website: https://ffmpeg.org/download.html")
            print(error_message)  # TODO how do errors get reported to the user?
            self.add_error(error_message)
            return False

        converted_file_path = None
        try:
            # Check and convert audio file if necessary
            converted_file_path = process_audio_file(
                file_path,
                target_format=self.target_format,
                target_sample_rate=self.target_sample_rate,
                target_bit_depth=self.target_bit_depth,
                target_channels=self.target_channels
            )

            self.dn_tracer.log_event(self.token, {
                DNTag.DNMsgStage.value: DNMsgStage.CLIENT_CONVERT_UPLOAD.value,
                DNTag.DNMsg.value: str(converted_file_path),
            })

        except Exception as e:
            self.dn_tracer.log_error(self.token, {
                DNTag.DNMsgStage.value: DNMsgStage.CLIENT_CONVERT_UPLOAD.value,
                DNTag.DNMsg.value: str(e),
            })
            self.add_error(str(e))

        try:
            file_url = await self.file_uploader.upload(converted_file_path,
                                                       os.path.splitext(converted_file_path)[1][1:])

            self.files.append({'name': os.path.basename(converted_file_path), 'url': file_url})

            self.dn_tracer.log_event(self.token, {
                DNTag.DNMsgStage.value: DNMsgStage.CLIENT_UPLOAD_ASSET.value,
                DNTag.DNMsg.value: file_url,
            })
        except Exception as e:
            self.dn_tracer.log_error(self.token, {
                DNTag.DNMsgStage.value: DNMsgStage.CLIENT_UPLOAD_ASSET.value,
                DNTag.DNMsg.value: str(e),
            })
            self.add_error(str(e))

        return True

    async def add_message(self, message):
        self.messages.append(message)
        return True

    async def add_log(self, log):
        self.logs += log
        return True

    def clear_outputs(self):
        """Clears the output attributes of the ResultsHandler instance."""
        self.message_id = None
        self.errors = []
        self.files = []
        self.logs = ""
        self.messages = []

    async def send(self):

        status = 'completed' if not self.errors else 'error'

        print("STATUS: " + status)

        data = {
            "response": {
                "files": self.files,
                "error": ", ".join(self.errors) if self.errors else None,
                "logs": self.logs,
                "message": ", ".join(self.messages) if self.messages else None,
                "status": status,
            }
        }

        if self.message_id:
            data["response"]["id"] = self.message_id

        send_msg = {
            'token': self.token,
            'type': 'results',
            'data': data
        }

        await self.websocket.send(json.dumps(send_msg))

        self.dn_tracer.log_event(self.token, {
            DNTag.DNMsgStage.value: DNMsgStage.CLIENT_SEND_RESULTS_MSG.value,
            DNTag.DNMsg.value: json.dumps(send_msg),
        })


def handle_the_results():
    return True



# test_results_handler.py
import pytest
from dawnet_client.output.results_handler import ResultsHandler

def test_clear_outputs():
    # Create an instance of ResultsHandler
    handler = ResultsHandler("websocket", "token")

    # Modify its attributes
    handler.message_id = "some_message_id"
    handler.errors = ["error1", "error2"]
    handler.files = ["file1", "file2"]
    handler.messages = ["message1", "message2"]

    # Call clear_outputs
    handler.clear_outputs()

    # Assert that the attributes are reset to their initial state
    assert handler.message_id is None
    assert handler.errors == []
    assert handler.files == []
    assert handler.messages == []



import os

from aiohttp import ClientSession
from .config import API_BASE_URL


class FileUploader:
    async def get_signed_url(self, filename, token) -> str:
        url = f"{API_BASE_URL}/api/hub/get_signed_url/?token={token}&filename={filename}"
        async with ClientSession() as session:
            async with session.get(url) as response:
                # TODO: Check if response is ok
                data = await response.json()
                return data['signed_url']


    async def upload_file_to_gcp(self, file_path, signed_url, file_type) -> bool:
        async with ClientSession() as session:
            with open(file_path, 'rb') as file:
                async with session.put(signed_url, data=file, headers={'Content-Type': file_type}) as response:
                    return response.status == 200


    async def upload(self, file_path, file_type) -> str:
        file_name = os.path.basename(file_path)
        file_url = f"https://storage.googleapis.com/byoc-file-transfer/{file_name}"
        signed_url = await self.get_signed_url(file_name, 'myToken')
        result = await self.upload_file_to_gcp(file_path, signed_url, file_type)

        if result:
            return file_url
        else:
            raise Exception("Failed to upload file to GCP")


from setuptools import setup, find_packages
import os
import subprocess

def is_ffmpeg_installed():
    try:
        subprocess.run(["ffmpeg", "-version"], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
        return True
    except FileNotFoundError:
        return False

ffmpeg_warning_msg = """
Warning: FFmpeg is not installed on your system, which is required for some functionalities of this package.
Please install FFmpeg using the following instructions:

- For macOS:
  Install FFmpeg using Homebrew: `brew install ffmpeg`

- For Linux (Debian/Ubuntu):
  Install FFmpeg using apt: `sudo apt-get install ffmpeg`

After installing, ensure that FFmpeg is in your PATH.
"""

if not is_ffmpeg_installed():
    print(ffmpeg_warning_msg)

setup(
    name='dawnet-client',
    version='0.1.1',
    packages=find_packages(),
    install_requires=[
        'websockets',
        'nest-asyncio',
        'sentry-sdk',
        'pydub',
        'librosa',
        'pytest-asyncio',
    ],
    python_requires='>=3.6',
    entry_points={
        'console_scripts': [
            'dawnet-client=dawnet_client.core:main',
        ],
    },
    # Additional metadata about your package
    author="Steve Hiehn",
    author_email="stevehiehn@gmail.com",
    description="DAWNet client enables remote execution of python code triggered from a DAW.",
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    url="https://dawnet.tools",
)



import os

# --------- PRODUCTION SETTINGS ----------------
SOCKET_IP = os.getenv('DN_CLIENT_SOCKET_IP', '35.223.141.253')
API_BASE_URL = os.getenv('DN_CLIENT_API_BASE_URL', '35.223.141.253:')
SENTRY_API_KEY = os.getenv('DN_CLIENT_SENTRY_API_KEY', "https://dbbf2855a707e0448957ad77111449c3@o4506379662131200.ingest.sentry.io/4506379670847488")
SOCKET_PORT = os.getenv('DN_CLIENT_SOCKET_PORT', '8765')
